library local_pubsub;

import 'dart:async';

import 'package:stack/stack.dart';

import '../../props/UikAction.dart';
//import 'package:ui_sdk/models/Action.dart';

//import 'package:flutter/material.dart';

/// [PubSub] manages its subscriptions.
/// You can subscribe and unsubscribe to topics, send them messages.
/// Example:
/// ```dart
/// PubSub pubsub = PubSub();
/// Subscription? sub = pubsub?.subscribe('topic');
/// sub?.stream?.listen((message) {
///   print(message);
/// });
///
/// pubsub.publish('topic', 'hello');
/// // console log - hello
/// pubsub.unsubscribe(sub);
/// ```
class PubSub {
  /// Map of topics and their subscribers.
  final _subscriptions = Map<String?, Set<Subscription?>?>();
  final stack = Stack();
  final screensSet = Set();

  /// Default name. If you want to use multiple pubsub you can set a name for each.
  String? name = 'Default';

  static final PubSub _singleton = PubSub._internal();

  factory PubSub() {
    return _singleton;
  }
  PubSub._internal();

  /// Sends messages to all subscribers of this topic.
  void publish(String? topic, UikAction uikAction) {
    if (!_subscriptions.containsKey(topic)) {
      return;
    }
    // print(topic);
    print(_subscriptions[topic]!.length);
    _subscriptions[topic]!.forEach((sub) {
      if (sub?.isCanceled == true) {
        throw Exception("This subscription canceled");
      }
      sub?.send(uikAction);
    });
  }

  /// Sends messages to all subscribers of several topics.
  void publishMany(Set<String?> topics, UikAction uikAction) {
    topics.forEach((topic) {
      publish(topic, uikAction);
    });
  }

  /// Sends messages to all subscribers of all topics.
  void publishAll(dynamic message) {
    _subscriptions.keys.forEach((topic) {
      publish(topic, message);
    });
  }

  /// Creates and stores a subscription to a topic.
  /// Return [Subscription].
  Subscription? subscribe(String? topic, dynamic? reference) {
    // if (screensSet.contains(reference)) return null;
    // screensSet.add(reference);
    Subscription? sub = Subscription({topic}, this.name, reference);

    // ignore: prefer_collection_literals
    _subscriptions.putIfAbsent(topic, () => Set<Subscription?>())!.add(sub);
    //  print(_subscriptions.length);
    //  print(_subscriptions[topic]!.length);
    var arr = [];
    _subscriptions[topic]!.forEach((element) {
      arr.add(element);
    });
    if (_subscriptions[topic]!.length == 2) {
      //print(arr[0] == arr[1]);
    }
    return sub;
  }

  /// Creates and stores a subscription to several topics.
  /// Return [Subscription].
  Subscription? subscribeToMany(Set<String?>? topics, dynamic? reference) {
    Subscription? sub = Subscription(topics, this.name, reference);
    topics!.forEach((topic) {
      _subscriptions.putIfAbsent(topic, () => Set<Subscription?>())?.add(sub);
    });
    return sub;
  }

  /// Deletes a subscription to a topic.
  //A   O_w   O_r
  //B   O_w   O-r

  //O_w  --> A,B
  //O_r  --> A,B

  //subArray B,B
  //actionList 0_r 0_w
  Future<void>? unsubscribe(Map<String?, Subscription?>? subArray) async {
    subArray?.forEach((key, value) {
      _subscriptions[key]?.remove(value);
    });
    //print(reference);
    //_subscription (string,set<subscription>)
    // for (int i = 0; i < actionList.length; i++) {
    //   String str = actionList[i];
    //  // _subscriptions[str]!
    // }
    // for (var i in _subscriptions.keys) {
    //   for (var j in _subscriptions[i]!) {
    //     if (j!.reference == reference) {
    //       _subscriptions[i]!.remove(j);
    //       break;
    //     }
    //   }
    // }
    // if (subscription?.pubsub != this.name) {
    //   throw Exception("This subscription doesn\'t belong to this pubsub");
    // }

    // subscription?.topics!.forEach((topic) {
    //   if (!_subscriptions.containsKey(topic)) {
    //     return;
    //   }

    //   print(_subscriptions[topic]!.length);
    //   _subscriptions[topic]!.remove(subscription);
    //   print(_subscriptions[topic]!.length);
    //   //removing the key
    //   if (_subscriptions[topic]!.isEmpty) {
    //     _subscriptions.remove(topic);
    //   }
    // });

    // await subscription?.cancel();
  }

  /// Deletes multiple subscriptions at a time.
  Future<void>? unsubscribeMany(List<Subscription?>? subscribers) {
    subscribers!.forEach((sub) async {
      //    await unsubscribe(sub, {});
    });
  }

  /// Deletes all subscriptions to the specified topic.
  Future<void>? unsubscribeAll(String? topic) async {
    if (!_subscriptions.containsKey(topic)) {
      return;
    }

    Set<Subscription?>? subs = Set.from(_subscriptions[topic]!);
    subs.forEach((sub) async {
      // await unsubscribe(sub, {});
    });
  }

  /// Return set of pubsub topics
  Set<String?>? getTopics() {
    return _subscriptions.keys.toSet();
  }
}

/// Pubsub subscription result.
///
/// [Subscription] can receive messages from pubsub and process them on its [stream].
/// Example:
/// ```dart
/// PubSub pubsub = PubSub();
/// Subscription? sub = pubsub?.subscribe('topic');
/// sub?.stream?.listen((message) {
///   print(message);
/// });
/// ```
class Subscription {
  /// Topics that are subscribed to.
  final Set<String?>? _topics;

  /// Pubsub name to check
  final String? _pubsub;
  final dynamic? reference;

  /// Controller for listening to messages.
  final StreamController<dynamic>? _controller = StreamController();

  /// Is subscription canceled
  bool? _isCanceled = false;

  Subscription(this._topics, this._pubsub, this.reference);

  Set<String?>? get topics => _topics;
  Stream<dynamic>? get stream => _controller?.stream;
  String? get pubsub => _pubsub;
  bool? get isCanceled => _isCanceled;

  /// PubSub uses this method to send messages to subscription.
  /// This method not intended to be manual called.
  void send(UikAction uikAction) {
    _controller?.add(uikAction);
    // print(uikAction);
  }

  /// PubSub uses this method to cancel subscription on it
  /// This method not intended to be manual called.
  /// dispose to free the resources
  Future<void>? cancel() async {
    await _controller?.close();
    _isCanceled = true;
  }
}
